//---------------------------------------------------------------------------
// Mpeg decoding is in a separate file due to its bigger complexity
//---------------------------------------------------------------------------
// Inspired from bbdmux (bbtool19_src.zip) by Brent Beyeler, beyeler@home.com
// Audio header analyse from FlaskMpeg by Alberto Vigata, ultraflask@yahoo.com
// Audio doc : http://www.dv.co.yu/mpgscript/mpeghdr.htm
// Trailers :
// - ftp://ftp.csst.net/csst/Dolby  (login:csst, passwd:csst)
// - http://www.digital-digest.com/dvd/downloads/trailers.html
// - http://www.cinenow.com/fr/vobtrailer.php3
// -----------------------------------------------------------
//      { MPEG_NORM_FILM,   352, 240, 1 },
//      { MPEG_NORM_PAL,    352, 288, 3 },
//      { MPEG_NORM_NTSC,   352, 240, 4 },
//      { MPEG_NORM_PAL_S,  480, 576, 3 },
//      { MPEG_NORM_NTSC_S, 480, 480, 4 },
//      { MPEG_NORM_OTHER, }

#include <vcl.h>						// ShowMesssage
#pragma hdrstop

#include <io.h>							// filelength
#include <stdio.h>						// fopen

#include "Main.h"						// MainForm			
#include "MpegInfo.h"
#include "Fileinfo.h"
#include "Multinfo.h"

extern struct FileInfo *pai;

//---------------------------------------------------------------------------

#pragma package(smart_init)

// -----------------------------------------------------------
// MPG info research
// -----------------------------------------------------------

#define MAXAUDIOMPEGSTREAM			10	// False, just to have a number

int ParseMpeg2 (wchar_t *wszFile)
{
	char	szTrack[10];
	bool	bFirstPSC = true,
    		bAudioStream[MAXAUDIOMPEGSTREAM],
    		bAC3Stream[MAXAUDIOMPEGSTREAM],
			bDTS,
			bLPCM,
            bEnd,
			bMpeg2,
            bMpeg1audio;
	int		as,
    		ar_mpg1[16] = {	0, 100, 67, 70, 76, 81, 84, 89, 94, 98, 103, 107, 112, 116, 120, 0   },
    		ar_mpg2[16] = { 0, 133, 178, 221, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
    		fr[9] = { 0,
				23976,		// 1 (23.976 fps) - Cinma = 24 fps
				24000,		// 2 (24.000 fps)
				25000,		// 3 (25.000 fps) - PAL
				29970,		// 4 (29.970 fps) - NTSC = 30 fps
				30000,		// 5 (30.000 fps)
				50000,		// 6 (50.000 fps) - PAL noninterlaced
				59940,		// 7 (59.940 fps) - NTSC noninterlaced
				60000,		// 8 (60.000 fps)
            },
			mpeg1_bitrate_table[3][15] = {
	{ 0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448 },
	{ 0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384 },
	{ 0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320 }},
			mpeg2_bitrate_table[3][15] = {
	{ 0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256, },
	{ 0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, },
	{ 0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, }};
    int    AC3Acmod[]={ 2, 1, 2, 3, 3, 4, 4, 5 };
	int    AC3SampleFreq[4]={48000,44100,32000,0};
	struct
	{
		WORD	bit_rate;
		WORD	frm_size[3];
	} AC3Framesize[] = {
      { 32  ,{64   ,69   ,96   } },
      { 32  ,{64   ,70   ,96   } },
      { 40  ,{80   ,87   ,120  } },
      { 40  ,{80   ,88   ,120  } },
      { 48  ,{96   ,104  ,144  } },
      { 48  ,{96   ,105  ,144  } },
      { 56  ,{112  ,121  ,168  } },
      { 56  ,{112  ,122  ,168  } },
      { 64  ,{128  ,139  ,192  } },
      { 64  ,{128  ,140  ,192  } },
      { 80  ,{160  ,174  ,240  } },
      { 80  ,{160  ,175  ,240  } },
      { 96  ,{192  ,208  ,288  } },
      { 96  ,{192  ,209  ,288  } },
      { 112 ,{224  ,243  ,336  } },
      { 112 ,{224  ,244  ,336  } },
      { 128 ,{256  ,278  ,384  } },
      { 128 ,{256  ,279  ,384  } },
      { 160 ,{320  ,348  ,480  } },
      { 160 ,{320  ,349  ,480  } },
      { 192 ,{384  ,417  ,576  } },
      { 192 ,{384  ,418  ,576  } },
      { 224 ,{448  ,487  ,672  } },
      { 224 ,{448  ,488  ,672  } },
      { 256 ,{512  ,557  ,768  } },
      { 256 ,{512  ,558  ,768  } },
      { 320 ,{640  ,696  ,960  } },
      { 320 ,{640  ,697  ,960  } },
      { 384 ,{768  ,835  ,1152 } },
      { 384 ,{768  ,836  ,1152 } },
      { 448 ,{896  ,975  ,1344 } },
      { 448 ,{896  ,976  ,1344 } },
      { 512 ,{1024 ,1114 ,1536 } },
      { 512 ,{1024 ,1115 ,1536 } },
      { 576 ,{1152 ,1253 ,1728 } },
      { 576 ,{1152 ,1254 ,1728 } },
      { 640 ,{1280 ,1393 ,1920 } },
      { 640 ,{1280 ,1394 ,1920 } }};
    unsigned char	stream_id;
    int		aspect_ratio,
    		n,
            idx,
    		a_type,
            a_val,
    		initTime;
	BYTE	bVal,
    		bVal2;
	WORD	code;
    DWORD	bitrate;
	DWORD	PES_header_data_length;
	BitReader br;
    bool	bInitTime,					// Is this information found ?
    		bVideo,
            bAudio;

	br.Open (wszFile);

	for (n=0 ; n<MAXAUDIOMPEGSTREAM ; n++)
    	bAudioStream[n] = bAC3Stream[n] = false;

	bLPCM = bDTS = bEnd = false;
	bInitTime = bVideo = bAudio = false;
	initTime = -1;
	do
    {
		code = br.getByte(2);
		while (code && !br.bEof)		// Go forward till obtaining 0x0000
        	code = (code << 8) | br.getByte();
		do	{
		bVal = br.getByte();
        } while (!bVal && !br.bEof);	// Skip all following 0
		if (br.bEof)	break;
       	if (bVal != 0x01)	continue;	// Not a Packet Start Code Prefix

		stream_id = br.getByte();		// We have a mpeg start code
       	code = 0x0100 | stream_id;
        switch (code)
        {
		case MPEG_PACK_SC:
        	if (bFirstPSC)
            {
                bVal = br.getBits (2);
                br.getBits (6);			// To be byte aligned
				bMpeg2 = (bVal == 1);
				strcpy (pai->szVC, !bMpeg2 ? "MPG1" : "MP2P");
				pai->dwFileHeadSize = br.ftell();
                bFirstPSC = false;
            }
            break;

/*		case MPEG_PROGRAM_END_CODE:
			br.bEof = true;				// Force exit
            break;
*/
		case MPEG_SEQUENCE_SC:			// Inside video stream
			if (pai->dwWidth)			// Already studied
				break;

			if (!*pai->szVC)			// Happens when no MPEG_PACK_START_CODE (XingCD)
				strcpy (pai->szVC, "MPG1");

        	pai->dwWidth = br.getBits (12);
        	pai->dwHeight = br.getBits (12);
			aspect_ratio = br.getBits (4);
			// 2007 01 : I use MediaInfo analyse in File_Mpegv.cpp
            if (!bMpeg2)				// Mpeg1
            {
				if (pai->dwHeight && ar_mpg1[aspect_ratio])
					pai->wAspRatio = pai->dwWidth / pai->dwHeight / ar_mpg1[aspect_ratio];
            }
            else						// Mpeg2
            {
/*				if ((aspect_ratio==1) &&  pai->dwHeigth)	// MediaInfo ? (ar_mpg2 is different)
					 pai->wAspRatio = pai->dwWidth / pai->dwHeight;
                else*/
                pai->wAspRatio = ar_mpg2[aspect_ratio];	// Ok for VOB
            }

			idx = br.getBits (4);
			pai->dwFps = fr[idx];
			idx = br.getBits (18);
            if (idx != 0x3FFFF) 		// 0x3FFFF if VBR Mpeg file
            {
	            bitrate = 400 * idx;	// 1152 for VCD, 2500 for SVCD, 3800 for DVD
				pai->dwVAvgBPS = bitrate / 8;
            }
			pai->wVCsupp = ftSUP;		// It is supported with correct player
			br.getBits (6);				// To be byte aligned
            bVideo = true;
            break;

		case MPEG_GOP_SC:				// Group-of-Pictures
			int	hour,
               	minute,
				second,
				frame,
				time;

			br.getBits (1);				// drop_frame
			hour = br.getBits (5);
			minute = br.getBits (6);
			br.getBits (1);				// time_code
			second = br.getBits (6);
			time = second + 60 * minute + 3600 * hour;
			frame = br.getBits (6);
			br.getBits (1);				// closed_gop
			br.getBits (1);				// broken links
			br.getBits (5);				// To be byte aligned

			if (initTime == -1)			//
				initTime = time;
        	else
				pai->dwFileDuration = pai->a[0].dwASec = pai->dwVSec = time + 1 - initTime;
 			break;

		case MPEG_USER_DATA_SC:
			if (!*pai->szFileInfo)		// No comment found yet
	        {
				char c, *p = pai->szFileInfo;
                int	n=0;

				do {
					c = br.getByte();
					*p++ = c;
                    n++;
	            } while (c && (n<100));
				*p++ = c;
        	    br.getBack();				// Return one step back
			}
			break;
		case MPEG_PRIVATE_STREAM_1:
			// Explanations on AC3 from ac3dec by Aaron Holtzman (included in FlaskMepeg source)
			DWORD	PES_packet_length;	// Data
			BYTE	substream_id;

			PES_packet_length = br.getByte(2);
			br.getByte(2);
			PES_header_data_length = br.getByte(1);
			for (n=0 ; n<(int)PES_header_data_length ; n++)
				br.getByte(1);
			substream_id = br.getByte(1);

			if ((substream_id >= SUBSTREAM_AC3_0)  && (substream_id <= SUBSTREAM_AC3_8))
            {
				WORD	sync;

				if (bAC3Stream[substream_id - SUBSTREAM_AC3_0] == true)	// Already studied
        	        break;
				bAC3Stream[substream_id - SUBSTREAM_AC3_0] = true;
				pai->wAStreamsNb++;
    	        as = pai->wAStreamsNb-1;
				if (as >= MAXAUDIOSTREAMS)
                	break;

				n = PES_packet_length - PES_header_data_length - 10;
				sync = br.getByte(1);
				do
        	    {
					sync = (sync << 8) + br.getByte(1);
					if (sync == AUDIO_AC3_SYNC_WORD)
                    	break;
                	n--;
	            }	while (n);
    	        if (!n)						// AC3 sync not found
	            	break;

				pai->a[as].wAC = 0x2000;			// AC3
				strcpy (pai->a[as].szACname, searchACname (pai->a[as].wAC));
				wsprintf (szTrack, " [0x%x]", substream_id);
                strcat (pai->a[as].szACname, szTrack);
				br.getByte(2);

				a_type = br.getBits (2);		// If 3, error
				pai->a[as].dwASps = AC3SampleFreq[a_type];

				a_val = br.getBits (6);
				pai->a[as].dwAAvgBPS = 1000 / 8 * AC3Framesize[a_val].bit_rate;
//				frame_size = AC3Framesize[a_val].frm_size[a_type];
//				bsi->bsid = br.getBits (5);	// AC-3 version number
//				bsi->bsmod = br.getBits (3);	// Audio service provided by the stream
  				br.getByte(1);
  				a_type = br.getBits (3);		// Audio coding mode
  				pai->a[as].wAChannels = AC3Acmod[a_type];	// Number of full bandwidth channels
				pai->a[as].wACsupp = ftSUP;	// It is supported with correct player

  				pai->dwFileHeadSize = br.ftell();
  				br.getBits (5);				// To be byte aligned
            }
			else if ((substream_id >= SUBSTREAM_DTS_0)  && (substream_id <= SUBSTREAM_DTS_8))
            {
				if (!bDTS)
                {
					if (!bOptions[OPT_Silent])
						ShowMessage (szErrorString[15]);
                    bDTS = true;

					pai->wAStreamsNb++;
    		        as = pai->wAStreamsNb-1;
					if (as >= MAXAUDIOSTREAMS)
            	    	break;
					pai->a[as].wAC = 0x0008;			// DTS
                }
            }
			else if ((substream_id >= SUBSTREAM_PCM_0)  && (substream_id <= SUBSTREAM_PCM_F))
            {
				if (!bLPCM)
                {
					if (!bOptions[OPT_Silent])
						ShowMessage (szErrorString[16]);
                    bLPCM = true;
                }
            }
 			break;

		case MPEG_AUDIO_C0_CODE:
		case MPEG_AUDIO_C1_CODE:
		case MPEG_AUDIO_C2_CODE:
			if (bAudioStream[stream_id - 0xc0] == true)	// Already studied
                break;
			if (bAC3Stream[0] == true)	// AC3 already found
       	        break;
			bAudioStream[stream_id - 0xc0] = true;
            as = pai->wAStreamsNb;
			if (as >= MAXAUDIOSTREAMS)
               	break;

			// I don't try to understand PES data ...

			PES_packet_length = br.getByte(2);
			if (!bMpeg2)
            {
 				PES_header_data_length = 7;
				while (br.getByte() == 0xFF)
					PES_header_data_length++;
   	            br.getByte();
   	            br.getByte(4);
   	            br.getByte();
			}
            else
            {
				br.getByte(2);
				PES_header_data_length = br.getByte();
				for (n=0; n<(int)PES_header_data_length; n++)
					br.getByte();
				substream_id = br.getByte();
            }

			n = PES_packet_length - PES_header_data_length - 10;
			bVal = br.getByte();
			while (n--)
            {
				bVal2 = br.getByte();
				if ((bVal == 0xFF) && ((bVal2&0xF0) == 0xF0))// AUDIO_MPEG_SYNC_WORD 0xFFFn found
                	break;
				bVal = bVal2;
            }
			if (!n)						// Not found
            	break;

			bMpeg1audio = (bVal2 & 0x0F)>>3;	// version, 1=Mpeg1, 0=Mpeg2
			a_type = 4-((bVal2>>1) & 0x03);	// layer (3 = layer I, 2 = layer II, 1 = layer III)
			pai->a[as].wAC = (a_type==1 ? 0x0050 : a_type==2 ? 0x0051 : a_type==3 ? 0x0053 : 0);
			strcpy (pai->a[as].szACname, searchACname (pai->a[as].wAC));
			wsprintf (szTrack, " [0x%x]", stream_id);
			strcat (pai->a[as].szACname, szTrack);

            a_val = br.getBits (4);		// bitrate_index
            if (a_val == 0xF)	break;	// Invalid value
			pai->a[as].dwAAvgBPS = 1000 / 8 * (bMpeg1audio ? mpeg1_bitrate_table : mpeg2_bitrate_table)[a_type-1][a_val];

			a_type = br.getBits (2);	// sampling_freq
			pai->a[as].dwASps = (a_type==0 ? 44100 : a_type==1 ? 48000 : a_type==2 ? 32000 : 0);
			if (!bMpeg1audio)
				pai->a[as].dwASps >>= 1;// Divide by 2

            br.getBits (2);				// padding + extension
			if (br.getBits (2) == 3)		// mode (0=stereo, 1=joint stereo, 2=dual channel, 3=mono)
					pai->a[as].wAChannels = 1;
			else	pai->a[as].wAChannels = 2;
            br.getBits (6);				// To be byte aligned : mode_ext + copyright + original + emphasis
			pai->a[as].wACsupp = ftSUP;	// It is supported with correct player

			// Valid audio stream
			pai->wAStreamsNb++;
        }
	// Stop at end of file or after 0x100000 bytes
    }	while (!br.bEof && (br.nBufRefill < 100));

	if (!br.bEof)
    {
		DWORD	dwStart;
		// Go to the EOF-0x90000 to read last gop and calculate time length

		dwStart = max(pai->dwFileRealSize - 90*BUFFER_SIZE, br.ftell ());
		br.fseek (dwStart, SEEK_SET);
		do
    	{
			code = br.getByte(2);
			while (code && !br.bEof)		// Go forward till obtaining 0x0000
        		code = (code << 8) | br.getByte();
			do	{
			bVal = br.getByte();
        	} while (!bVal && !br.bEof);	// Skip all following 0
			if (br.bEof)	break;
    	   	if (bVal != 0x01)	continue;	// Not a Packet Start Code Prefix

			code = 0x0100 | br.getByte();
			if (code == MPEG_GOP_SC)// Group-of-Pictures
            {
				int	hour, minute, second, frame, time;

				br.getBits (1);				// drop_frame
				hour = br.getBits (5);
				minute = br.getBits (6);
				br.getBits (1);				// time_code
				second = br.getBits (6);
				time = second + 60 * minute + 3600 * hour;
				frame = br.getBits (6);
				br.getBits (1);				// closed_gop
				br.getBits (1);				// broken links
				br.getBits (5);				// To be byte aligned

				pai->dwFileDuration = pai->a[0].dwASec = pai->dwVSec = time + 1 - initTime;
            }
	    }	while (!br.bEof);
	}

	br.Close ();
    return 0;
}

