//---------------------------------------------------------------------------
// Multimedia Tools
// Another good source for codec info : http://avifile.sourceforge.net/
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#pragma warn -pin						// Suppr warning "Init is partially bracketed"

#include "dshow.h"						// DirectShow
#include <stdio.h>						// fopen
#include <vfw.h>						// ICINFO

#include "Main.h"
#include "Multinfo.h"

// The 3 main aspect ratio existing
// -----------------------------------------------------------

struct ASPECTRATIO ara[MAXARA]
=
{
"5:4", 125, 111, 129,
"4:3", 133, 130, 155,
"16:9", 178, 156, 200,
"2.21:1", 221, 201, 241,
"?", -1, -1, -1,					// This text is translated in MainForm
};

char *searchAspRatio (int aratio)
{
	static char szAspRatio[8];
	int	n=0;

	while (TRUE)
       	if ((aratio >= ara[n].min100) && (aratio < ara[n].max100))
			return ara[n].szAra;
       	else if (ara[n].num100 == -1)
        {
			n = aratio/100;
			wsprintf (szAspRatio, "%d.%d:1", n, aratio-n*100);
			return szAspRatio;
        }
        else	n++;
}

// Video Codec tools
// -----------------------------------------------------------

int searchvc (char *szFCC)
{
	int	n = vcd.size();

	while (--n)
		if (!strcmpi (vcd[n].szFCC, szFCC))	// Case insensitive search
			break;
	return n;
}

char *searchVCname (char *szFCC)
{
	return vcd[searchvc (szFCC)].szName;
}

char *searchVCurl (char *szFCC)
{
	return vcd[searchvc (szFCC)].szURL;
}

char szFileExtension[200];
struct VIDEOCODEC	vc1;
struct AUDIOCODEC	ac1;
vector<struct VIDEOCODEC>	vcd;
vector<struct AUDIOCODEC>	acd;

int InitCodecs (char *szInstallDir)
{
	BOOL	bVideo;
	char	szBuf[255],
    		szFile[MAXPATH],
            *p, *q;
	FILE	*fh;

	wsprintf (szFile, "%s\\codecs.xml", szInstallDir);
	fh = fopen (szFile, "rt");
	if (fh == NULL)
	{	strcpy (szErr, szErrorString[17]); return -1;	}

	while (fgets (szBuf, 255, fh))
	{
		if (strstr (szBuf, "Video"))
			bVideo = true;
		if (strstr (szBuf, "Audio"))
			bVideo = false;
		if (strstr (szBuf, "<FileExtension>"))
        {
			strcpy (szFileExtension, strchr (szBuf, '>')+1);
			*(strchr (szFileExtension, '<')) = '\0';
        	continue;
        }
		if (!strstr (szBuf, "<Codec>"))
        	continue;

		p = strstr (szBuf, "<Code>")+6;
		q = strstr (p, "</Code>");
		*q = 0;
		if (bVideo)
				strcpy (vc1.szFCC, p);
		else	sscanf (p, "%x", &ac1.code);

		p = strstr (q+1, "<Name>")+6;
		q = strstr (p, "</Name>");
		*q = 0;
		strcpy (bVideo ? vc1.szName : ac1.szName, p);

		p = strstr (q+1, "<Url>")+5;
		q = strstr (p, "</Url>");
		*q = 0;
		strcpy (bVideo ? vc1.szURL : ac1.szURL, p);

		if (bVideo)
				vcd.push_back (vc1);
		else	acd.push_back (ac1);
	}
	fclose (fh);
    return 0;
}

int searchac (int code)
{
	int	n = acd.size();

	while (--n)
		if (acd[n].code == code)
			break;
	return n;
}

char *searchACname (int code)
{
	return acd[searchac (code)].szName;
}

char *searchACurl (int code)
{
	return acd[searchac (code)].szURL;
}

//---------------------------------

// Not inside SuystemInfo because of access problem from callback function
struct CodecInfo ci;
vector<CodecInfo>	sys_vc;				// System Video Codec/Filter
vector<CodecInfo>	sys_ac;				// System Audio Codec/Filter
vector<CodecInfo>	sys_ds;   			// System DirectShow Filter
vector<CodecInfo>	sys_dmo;   			// System DirectX Media Object
BOOL bCIinit=FALSE;						// Is CodecInfo initialized ?

// Enumeration of VFW audio codecs
// HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\AudioCompressionManager\DriverCache

BOOL	bPCM;                       // Is PCM already displayed (because a lot of drivers handle it)

BOOL CALLBACK SystemInfo::acmDriverEnumCallback (HACMDRIVERID hadid, DWORD dwInstance, DWORD fdwSupport)
{
	HACMDRIVER	had;
	ACMDRIVERDETAILS	add;
	ACMFORMATTAGDETAILS	aftd;
	int		v;

	memset (&add, 0, sizeof (ACMDRIVERDETAILS));
	add.cbStruct = sizeof (ACMDRIVERDETAILS);
	acmDriverDetails (hadid, &add, 0);

	if (acmDriverOpen (&had, hadid, 0))
		return TRUE;					// Error, but go to next
	v = add.vdwDriver;
	wsprintf (ci.szVers, "%d.%d .%d", v>>24, (v & 0xFF0000)>>16, (v & 0xFFFF));
	strcpy (ci.szComp, add.szShortName);

	for (DWORD i=0 ; i<add.cFormatTags ; i++)
	{
		memset (&aftd, 0, sizeof (ACMFORMATTAGDETAILS));
		aftd.cbStruct = sizeof (ACMFORMATTAGDETAILS);
		aftd.dwFormatTagIndex = i;
		if (acmFormatTagDetails (had, &aftd, ACM_FORMATTAGDETAILSF_INDEX))						// Error
			break;

		if (aftd.dwFormatTag == 1)		// PCM
			if (!bPCM)	bPCM = TRUE;
			else		continue;		// Already displayed, jump to next

		wsprintf (ci.szId, "0x%04X", aftd.dwFormatTag);
        ci.dwFormatTag = aftd.dwFormatTag;
		strcpy (ci.szName, aftd.szFormatTag);
        sys_ac.push_back (ci);
	}

	acmDriverClose (had, 0);
	return TRUE;						// Finished with this driver, go to next
}

// Initialize structure CodecInfo

void __fastcall SystemInfo::InitCodecInfo ()
{
	// Init ci
	strcpy (ci.szId, "");
	strcpy (ci.szName, "");
	strcpy (ci.szFile, "");
	strcpy (ci.szVers, "");
	strcpy (ci.szComp, "");
}

char * __fastcall SystemInfo::getErrorText(HRESULT hr)
{
	AMGetErrorText(hr, szErr, SZERR); return szErr;
};

// Read version info from a file

#define MAXVIDATA	4096

int __fastcall SystemInfo::readVersion ()
{
	char	szBuf[200], *p,
			szQuery[100], *pQuery,
			viData[MAXVIDATA];
	DWORD	dwSize,
			handle;
	UINT	wSize;

	if (!*ci.szFile)
    	return -1;

	// Retrieving version info for the driver

	dwSize = GetFileVersionInfoSize (ci.szFile, &handle);
	if (dwSize)							// VersionInfo exists
	{
		if (dwSize >MAXVIDATA)	dwSize = MAXVIDATA;
		GetFileVersionInfo (ci.szFile, 0, dwSize, viData);
		wSize = (UINT) dwSize;
		strcpy (szQuery, "\\StringFileInfo\\");
		if (VerQueryValue (viData, "\\VarFileInfo\\Translation", &((void *)p), &wSize))
		{
			wsprintf (szBuf, "%04X%04X", *(SHORT *)p, *(SHORT *)(p+2));
			strcat (szQuery, szBuf);
		}
		else	strcat (szQuery, "040904E4");	// US language
		strcat (szQuery, "\\");
		pQuery = szQuery + strlen (szQuery);
		// Small test to verify if Translation seems correct
		strcpy (pQuery, "FileVersion");
		if (!VerQueryValue (viData, szQuery, &((void *)p), &wSize))
			strcpy (pQuery-9, "040904E4\\");	// US language

		if (!ci.szId[4])			// If Video codec, we prefer File description than mine
		{
			strcpy (pQuery, "FileDescription");
			if (VerQueryValue (viData, szQuery, &((void *)p), &wSize))
				strcpy (ci.szName, p);
        }
		strcpy (pQuery, "FileVersion");
		if (VerQueryValue (viData, szQuery, &((void *)p), &wSize))
        {
			if (!p[1])
					strcat (ci.szVers, WideCharToString ((wchar_t *)p).c_str());
			else	strcpy (ci.szVers, p);
        }

		strcpy (pQuery, "CompanyName");
		if (VerQueryValue (viData, szQuery, &((void *)p), &wSize))
				strcpy (ci.szComp, p);
		else	strcpy (ci.szComp, "No company info");
	}
	else
		strcpy (ci.szComp, szOtherString[12]);

    return 0;
}

// Init SystemInfo

__fastcall SystemInfo::SystemInfo ()
{
    char	szFile[MAXPATH];

	if (bCIinit)						// Already initialized
    	return;

	// First, we enumerate VFW Video codecs
    //----------------------------------------------------------------------

	InitCodecInfo ();
	ICINFO	icinfo;
    int		count=0;

	while (ICInfo (ICTYPE_VIDEO, count++, &icinfo))
    {
		*(DWORD *)ci.szId = (icinfo.fccHandler);
   	    ci.szId[4] = '\0';
		strcpy (ci.szName, searchVCname ((char *)&icinfo.fccHandler));

		strcpy(szFile, WideCharToString (icinfo.szDriver).c_str());
        ci.bDisabled = (*szFile == ':');
		GetSystemDirectory (ci.szFile, MAXPATH-1);
		strcat (ci.szFile, "\\");
		strcat (ci.szFile, ci.bDisabled ? szFile+1 : szFile);

		readVersion ();
        sys_vc.push_back (ci);
    }

	// Second, we enumerate VFW AUdio codecs
    //----------------------------------------------------------------------

	InitCodecInfo ();
	bPCM = FALSE;
	acmDriverEnum ((ACMDRIVERENUMCB)&SystemInfo::acmDriverEnumCallback, 0, 0);
    EnumAudioCodec();

	// Third, we enumerate DX filters
    //----------------------------------------------------------------------

	//  DirectShow-specific initialization code
	HRESULT hr;
	ICreateDevEnum *pSysDevEnum;
	IEnumMoniker *pEnumCat;

	// Create the System Device Enumerator.
	pSysDevEnum = NULL;
	hr = CoCreateInstance(CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC_SERVER,
		IID_ICreateDevEnum, (void **)&pSysDevEnum);
	if (hr == S_OK)
    {
		// Obtain a class enumerator for the DirectShow filters category.
		pEnumCat = NULL;
		hr = pSysDevEnum->CreateClassEnumerator(CLSID_LegacyAmFilterCategory, &pEnumCat, 0);
        if (hr == S_OK)
        {
	    	while (DXnext (pEnumCat) == S_OK)
				sys_ds.push_back (ci);
			pEnumCat->Release();
        }
	    else	getErrorText (hr);
		pSysDevEnum->Release();
	}
    else	getErrorText (hr);
    bCIinit = TRUE;

	// Fourth, we enumerate DMOs
    //----------------------------------------------------------------------

	if (!bOptions[OPT_DisableDMO])
    {
	    IEnumDMO *pEnum = NULL;

		// HKCR\DirectShow\MediaObjects
    	hr = DMOEnum (GUID_NULL, DMO_ENUMF_INCLUDE_KEYED, 0, NULL, 0, NULL, &pEnum);
	    if (hr == S_OK)
    	{
    		while (DMOnext (pEnum) == S_OK)
				sys_dmo.push_back (ci);
			pEnum->Release();
		}
	    else	getErrorText (hr);
    }
}

// Enumerate the DirectShow filters (DirectX), return S_OK if Next ok

int __fastcall SystemInfo::DXnext (IEnumMoniker *pEnumCat)
{
	bool	bFound = false;

	InitCodecInfo ();

	while (!bFound)
    {
		IMoniker *pMoniker = NULL;
		IPropertyBag *pPropBag = NULL;
//		IBaseFilter *pFilter = NULL;
		VARIANT	varFilterClsid = {0};
		VARIANT	varFilterName = {0};
		TCHAR	szKey[512];
		DWORD	dwSize;
        HRESULT	hr;
		HKEY	hkeyFilter = 0;
		int 	rc = 0;
//		CLSID	clsid;

		// Get next moniker
		hr = pEnumCat->Next(1, &pMoniker, NULL);
		if (hr != S_OK)	return hr;	// End or error

		hr = pMoniker->BindToStorage(0, 0, IID_IPropertyBag, (void **)&pPropBag);
		if FAILED(hr)	goto Clean;

		// Read CLSID string
		varFilterClsid.vt = VT_BSTR;
		hr = pPropBag->Read(L"CLSID", &varFilterClsid, 0);
		if FAILED(hr)	goto Clean;
		strcpy (ci.szId, WideCharToString(varFilterClsid.bstrVal).c_str());
        if (ci.szId[0]==':')
        {
        	ci.bDisabled = true;			// Codec is currently disabled
            strcpy(ci.szId, ci.szId+1);
        }
        else	ci.bDisabled = false;
//		CLSIDFromString(varFilterClsid.bstrVal, &clsid);
//		if ((clsid == CLSID_AVICo) ||		// AVI Compressor Filter
//			(clsid == CLSID_ACMWrapper))	// ACM Wrapper
//           goto Clean;

		// Retrieve the friendly name of the filter
		varFilterName.vt = VT_BSTR;
		hr = pPropBag->Read(L"FriendlyName", &varFilterName, 0);
		if FAILED(hr)	goto Clean;
		strcpy (ci.szName, WideCharToString(varFilterName.bstrVal).c_str());

		// Open the CLSID key that contains information about the filter
		dwSize = MAXCODFILE;
		wsprintf (szKey, TEXT("CLSID\\%s\\InprocServer32\0"), ci.szId);
		rc = RegOpenKey(HKEY_CLASSES_ROOT, szKey, &hkeyFilter);
		if (rc == ERROR_SUCCESS)
			rc = RegQueryValueEx(hkeyFilter, NULL, NULL, NULL, ci.szFile, &dwSize);
		rc = RegCloseKey(hkeyFilter);

		// Open the CLSID key that contains information about the filter
/*		dwSize = MAXCODCOMP;
		wsprintf (szKey, TEXT("Software\\Classes\\CLSID\\%s"), ci.szId);
		rc = RegOpenKey(HKEY_LOCAL_MACHINE, szKey, &hkeyFilter);
		if (rc == ERROR_SUCCESS)
			rc = RegQueryValueEx(hkeyFilter, NULL, NULL, NULL, ci.szComp, &dwSize);
		rc = RegCloseKey(hkeyFilter);

		hr = pMoniker->BindToObject(NULL, NULL, IID_IBaseFilter,
                                       (void**)&pFilter);
		if (SUCCEEDED(hr))
		{
			LPWSTR	VendorInfo=NULL;
	        FILTER_INFO info;

			if(SUCCEEDED(pFilter->QueryFilterInfo(&info)))
				strcpy (ci.szComp, WideCharToString(info.achName).c_str());

			pFilter->QueryVendorInfo (&VendorInfo);
            if (VendorInfo)
            {
				strcpy (ci.szComp, WideCharToString(VendorInfo).c_str());
				CoTaskMemFree (VendorInfo);
            }
		    pFilter->Release();
		}*/

        bFound = true;
		// Clean up.
Clean:;
		if (hr)
			getErrorText (hr);
		if (varFilterClsid.bstrVal)
			SysFreeString (varFilterClsid.bstrVal);
		if (varFilterName.bstrVal)
			SysFreeString (varFilterName.bstrVal);
		if (pPropBag)
			pPropBag->Release();
		pMoniker->Release();
    }

	readVersion ();
	return S_OK;
}

// Enumerate the DirectX Media Object, return S_OK if Next ok

int __fastcall SystemInfo::DMOnext (IEnumDMO *pEnumCat)
{
	InitCodecInfo ();

	TCHAR	szKey[512];
	DWORD	dwSize;
	HRESULT	hr;
	HKEY	hkeyFilter = 0;
	int 	rc = 0;

	CLSID	clsid;
	WCHAR	*wszCLSID, *wszName;

    // Get next moniker
	hr = pEnumCat->Next(1, &clsid, &wszName, NULL);
	if (hr != S_OK)	return hr;	// End or error

	StringFromCLSID (clsid, &wszCLSID);
	strcpy (ci.szId, WideCharToString(wszCLSID).c_str());
	strcpy (ci.szName, WideCharToString(wszName).c_str());

	// Open the CLSID key that contains information about the filter
	dwSize = MAXCODFILE;
	wsprintf (szKey, TEXT("CLSID\\%s\\InprocServer32\0"), ci.szId);
	rc = RegOpenKey(HKEY_CLASSES_ROOT, szKey, &hkeyFilter);
	if (rc == ERROR_SUCCESS)
		rc = RegQueryValueEx(hkeyFilter, NULL, NULL, NULL, ci.szFile, &dwSize);
	rc = RegCloseKey(hkeyFilter);

/* Disabled because its create an ntdll exception (strange ...)
	if (wszCLSID)
		SysFreeString (wszCLSID);
	if (wszName)
		SysFreeString (wszName);*/

	readVersion ();
	return S_OK;
}

#define KEYSIZE	100

void __fastcall SystemInfo::EnumAudioCodec()
{
	TCHAR	szKey[KEYSIZE], szSubKey[KEYSIZE];
	HKEY	hkey = 0,
		    hsubkey = 0;
	int 	rc = 0;
    char	szBuf[200],
    		szFile[MAXCODFILE];
	DWORD	dwSubKeySize = KEYSIZE,
			dwIndex = 0,
            dwType,
            dwFormatTags,
            dwSize,
            *p;
	FILETIME	ftLastWriteTime;

	// Open the CLSID key that contains information about the audio VFW codec
	rc = RegOpenKey(HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\AudioCompressionManager\\DriverCache\0", &hkey);
	if (rc != ERROR_SUCCESS)
    	return;

	while (RegEnumKeyEx(hkey, dwIndex++, szSubKey, &dwSubKeySize, NULL, NULL, NULL, &ftLastWriteTime) == ERROR_SUCCESS)
    {
		dwSubKeySize = KEYSIZE;    
		wsprintf(szKey, TEXT("SOFTWARE\\Microsoft\\AudioCompressionManager\\DriverCache\\%s\0"), szSubKey);
		rc = RegOpenKey(HKEY_LOCAL_MACHINE, szKey, &hsubkey);
		dwSize = 4;
		rc = RegQueryValueEx(hsubkey, "cFormatTags", 0, &dwType, szBuf, &dwSize);
        dwFormatTags = *(DWORD *)szBuf;
		dwSize = 200;
		rc = RegQueryValueEx(hsubkey, "aFormatTagCache", 0, &dwType, szBuf, &dwSize);
		rc = RegCloseKey(hsubkey);

		p = (DWORD *)szBuf;
		while (dwFormatTags--)
        {
            dwType = *p;
			p += 2;
            if (dwType == 1)
            	continue;

			for (int idx=0 ; idx< sys_ac.size() ; idx++)
            {
            	ci = sys_ac[idx];
                if ((ci.dwFormatTag == dwType) && !strchr(ci.szId, '.'))
				{
					strcat(ci.szId, ".");
                    strcat(ci.szId, strchr(szSubKey, '.')+1);
					dwSize = MAXCODFILE;
					rc = RegOpenKey(HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Drivers32\0", &hsubkey);
					rc = RegQueryValueEx(hsubkey, szSubKey, 0, &dwType, szFile, &dwSize);
                    if (strchr(szFile, '\\'))
                    	strcpy(ci.szFile, szFile);
					else
                    {
                    	GetSystemDirectory(ci.szFile, MAXCODFILE);
                        strcat(ci.szFile, "\\");
                        strcat(ci.szFile, szFile);
                    }
					rc = RegCloseKey(hsubkey);
                    sys_ac[idx] = ci;
					break;
                }
            }
        }
	}
	rc = RegCloseKey(hkey);
}



// Class BitReader
// -----------------------------------------------------------

__fastcall BitReader::BitReader (FILE *FileHandle)
{
	try {	buffer = new char [BUFFER_SIZE];	}
	catch (...)
	{	ShowMessage (szErrorString[0]);		}

	nBufCount = 0;
	nBufRefill = -1;
	bEof = false;

	fh = FileHandle;
	if (fh)
    {
		if (!refill_buffer ())
		{	strcpy (szErr, szErrorString[14]);	}
    }
}

__fastcall BitReader::~BitReader ()
{
	delete buffer;
}

int __fastcall BitReader::Open (wchar_t *wszFile)
{
	fh = _wfopen (wszFile, L"rb");
	if (fh == NULL)
	{	strcpy (szErr, szErrorString[7]);	return -1;	}
	if (!refill_buffer ())
	{	strcpy (szErr, szErrorString[14]);	return -2;	}
	return 0;
}

void __fastcall BitReader::Close ()
{
	fclose (fh);
}

// Fill buffer, return true if ok, false if end of file or problem

bool __fastcall BitReader::refill_buffer ()
{
	nByteIdx = 0;
	nBitIdx = 8;
	nBufRefill++;
	nBufCount = fread (buffer, 1, BUFFER_SIZE, fh);
    if (!nBufCount)						// 0 bytes read -> we reach the end of file
	    bEof = feof (fh) || ferror (fh);
    return !bEof;
}

// Read 1 bit from the bit stream
// On error returns 0 + eof = true
// -----------------------------------------------------------

DWORD __fastcall BitReader::get1bit ()
{
	DWORD bit;

	if (bEof)	return 0;				// Cannot read anything more

	bit = (buffer[nByteIdx] >> (nBitIdx-1)) & 0x1;
	nBitIdx--;
	if (!nBitIdx)
	{
		nBitIdx = 8;
		nByteIdx++;
		if (nByteIdx == nBufCount)
			refill_buffer ();
	}
	return bit;
}

// Read N bit from the bit stream (max 32)
// -----------------------------------------------------------

DWORD __fastcall BitReader::getBits (int n)
{
	DWORD	val = 0;
	int		i = n;

	if ((bEof) || (n>32))	return 0;

	// Optimization in case we are on byte boundary and want to read multiple of bytes!
	if ((nBitIdx == 8) && !(n & 7))
	{
		i = n >> 3;
		while (i > 0)
		{
			val = (val << 8) | (DWORD) buffer[nByteIdx];
			nByteIdx++;
			if (nByteIdx == nBufCount)
				if (!refill_buffer ())
                	break;;
			i--;
		}
    }
	else while (i > 0)
	{
		val = (val << 1) | (DWORD) get1bit ();
		i--;
    }

	return val;
}

// Read 1 to 4 "complete" byte(s) from the bit stream
// -----------------------------------------------------------

DWORD __fastcall BitReader::getByte (int n)
{
	DWORD	val = 0;

	if ((bEof) || (n>= 4))	return 0;
	if (nBitIdx != 8)	getBits (nBitIdx);	// Align to next byte

	for (int i=0 ; i<n ; i++)
	{
		val = (val << 8) | (DWORD) buffer[nByteIdx++];
		if (nByteIdx == nBufCount)
			if (!refill_buffer ())
               	return 0;
	}
	return val;
}

// Return back for n char
// If inside the buffer, nothing to do, else we go back and refill previous buffer

int __fastcall BitReader::getBack (int n)
{
	if (nByteIdx < n)
    {
		::fseek (fh, -2*BUFFER_SIZE, SEEK_CUR);
		nBufCount = fread (buffer, 1, BUFFER_SIZE, fh);
		nBufRefill--;
		nByteIdx += BUFFER_SIZE-n;
    }
    else
    {
		nByteIdx -= n;
    }
    return 0;
}

// Go to a special place and refill buffer
// Warning, this fseek change the file position due to buffer refilling

int __fastcall BitReader::fseek (long offset, int whence)
{
	int	ret;

	if (whence==SEEK_SET)
    {
		nBufRefill = offset/BUFFER_SIZE;
		nByteIdx = offset%BUFFER_SIZE;
		nBitIdx = 8;
	    ret = ::fseek (fh, nBufRefill*BUFFER_SIZE, whence);
		nBufCount = fread (buffer, 1, BUFFER_SIZE, fh);
	    if (!nBufCount)						// 0 bytes read -> we reach the end of file
		    bEof = feof (fh) || ferror (fh);
    }
	else
    {
	    ret = ::fseek (fh, offset, whence);
    	if (!ret)							// 0 means operation successful
			refill_buffer();
    }
	return ret;
}

// Return file pointer

int __fastcall BitReader::ftell ()
{
	return ::ftell(fh);
}

